# coding: utf-8
# 2020/1/13 @ tongshiwei
import json
import os
import uuid
import tempfile
import warnings
import functools
import queue
import threading
import logging
from longling import wf_open, path_append, loading

__all__ = ["BaseIter", "LoopIter", "AsyncLoopIter", "AsyncIter", "CacheAsyncLoopIter", "iterwrap"]


class Register(dict):  # pragma: no cover
    def add(self, cls):
        if cls.__name__ in self:
            warnings.warn("key %s has already existed, which will be overridden" % cls.__name__)
        self[cls.__name__] = cls
        return cls


register = Register()


@register.add
class BaseIter(object):
    def __init__(self, src, length=None, *args, **kwargs):
        self._reset = src
        self._data = None
        self._length = length
        self._count = 0
        self.init()

    def __next__(self):
        try:
            self._count += 1
            return next(self._data)
        except StopIteration:
            self._count -= 1
            self._set_length()
            raise StopIteration

    def __iter__(self):
        return self

    def __len__(self):
        if self._length is None:
            raise TypeError("length is unknown")
        return self._length

    def init(self):
        self.reset()

    def reset(self):
        _data = self._reset() if callable(self._reset) else self._reset
        try:
            self._length = len(_data)
        except TypeError:
            self._set_length()
        self._data = iter(_data)

    def _set_length(self):
        if self._length is None and self._count > 0:
            self._length = self._count

    @classmethod
    def wrap(cls, f):
        @functools.wraps(f)
        def _f(*args, **kwargs):
            return cls(functools.partial(f, *args, **kwargs))

        return _f


@register.add
class LoopIter(BaseIter):
    def __init__(self, src, length=None, *args, **kwargs):
        super(LoopIter, self).__init__(src, length)

    def __next__(self):
        try:
            self._count += 1
            return next(self._data)
        except StopIteration:
            self._count -= 1
            self.reset()
            raise StopIteration


@register.add
class AsyncLoopIter(LoopIter):
    def __init__(self, src, prefetch=True, tank_size=8, timeout=None):
        self.prefetch = prefetch
        self.thread = None
        self._size = tank_size
        self.queue = queue.Queue(self._size)
        self._stop = False
        self._timeout = timeout
        super(AsyncLoopIter, self).__init__(src)

    def reset(self):
        super(AsyncLoopIter, self).reset()
        if self.thread is not None:
            self.thread.join()
        self._stop = False
        if self.prefetch:
            self.thread = threading.Thread(target=self.produce, daemon=True)
            self.thread.start()

    def __next__(self):
        if not self.prefetch:
            self.produce(False)
        if self.queue is not None:
            item = self.queue.get()
        else:
            raise StopIteration
        if isinstance(item, Exception):
            if isinstance(item, StopIteration):
                self.reset()
                raise StopIteration
            else:  # pragma: no cover
                raise item
        else:
            return item

    def produce(self, daemon=True):
        try:
            for _data in self._data:
                self._count += 1
                self.queue.put(_data)
                if daemon is False:
                    return
            raise StopIteration
        except StopIteration as e:
            if not self._stop:
                self.queue.put(e)
                self._stop = True

        except Exception as e:  # pragma: no cover
            if not self._stop:
                self._stop = True
                self.queue.put(e)


@register.add
class AsyncIter(AsyncLoopIter):
    def init(self):
        super(AsyncIter, self).reset()

    def reset(self):
        self.thread = None
        self.queue = None
        self._set_length()


@register.add
class CacheAsyncLoopIter(AsyncLoopIter):
    def __init__(self, src, cache_file=None, rerun=True, prefetch=True, tank_size=8, timeout=None):
        if cache_file is None:
            cache_dir = tempfile.mkdtemp(str(uuid.uuid4()))
            filename = str(uuid.uuid4())
            self.cache_file = path_append(cache_dir, filename)
        else:
            self.cache_file = cache_file

        self.cache_queue = None
        self.cache_thread = None

        if os.path.exists(self.cache_file) and not rerun:
            # 从已有数据中进行装载
            def src():
                return loading(self.cache_file, "json")

            self._cache_stop = True
        else:
            # 重新生成数据
            self.cache_queue = queue.Queue(tank_size)
            self.cache_thread = threading.Thread(target=self.cached, daemon=False)
            self.cache_thread.start()
            self._cache_stop = False
        super(CacheAsyncLoopIter, self).__init__(src, prefetch, tank_size, timeout)

    def init(self):
        super(CacheAsyncLoopIter, self).reset()

    def reset(self):
        self._reset = lambda: loading(self.cache_file, "json")
        if self.cache_thread is not None:
            self.cache_thread.join()
            self.cache_thread = None
            self.cache_queue = None
        super(CacheAsyncLoopIter, self).reset()

    def __next__(self):
        if not self.prefetch:
            self.produce(False)
        item = self.queue.get()
        if self.cache_queue is not None:
            self.cache_queue.put(item)
        if isinstance(item, Exception):
            if isinstance(item, StopIteration):
                self.reset()
                raise StopIteration
            else:  # pragma: no cover
                raise item
        else:
            return item

    def cached(self):
        assert self.cache_queue is not None

        with wf_open(self.cache_file, mode="w") as wf:
            while True:
                data = self.cache_queue.get()
                if isinstance(data, StopIteration):
                    break
                print(json.dumps(data), file=wf)


def iterwrap(itertype: str = "AsyncLoopIter", *args, **kwargs):
    if itertype not in register:
        raise TypeError("itertype %s is unknown, the available type are %s" % (itertype, ", ".join(register)))

    def _f1(f):
        @functools.wraps(f)
        def _f2(*fargs, **fkwargs):
            return register[itertype](functools.partial(f, *fargs, **fkwargs), *args, **kwargs)

        return _f2

    return _f1



# coding: utf-8
# 2020/1/2 @ tongshiwei

import pathlib
from tqdm import tqdm
from longling import wf_open, rf_open, PATH_TYPE
import csv
import json

__all__ = ["csv2json", "json2csv", "load_csv", "load_json", "loading"]


def csv2json(src, tar, delimiter=",", **kwargs):
    with wf_open(tar) as wf:
        for line in tqdm(load_csv(src, delimiter=delimiter, **kwargs), "csv2json: %s --> %s" % (src, tar)):
            print(json.dumps(line), file=wf)


def json2csv(src, tar, delimiter=",", **kwargs):
    with wf_open(tar) as wf:
        csv_writer = None
        for line in tqdm(load_json(src), "json2csv: %s --> %s" % (src, tar)):
            if csv_writer is None:
                csv_writer = csv.DictWriter(wf, line.keys(), delimiter=delimiter, **kwargs)
                csv_writer.writeheader()
            csv_writer.writerow(line)


def load_file(src: PATH_TYPE):
    with rf_open(src) as f:
        for line in f:
            yield line


def load_csv(src: PATH_TYPE, delimiter=",", **kwargs):
    """read the dict from csv"""
    with rf_open(src) as f:
        field_names = [i for i in csv.reader([f.readline()], delimiter=delimiter, **kwargs)][0]

        for line in csv.DictReader(f, field_names, delimiter=delimiter, **kwargs):
            yield line


def load_json(src: PATH_TYPE):
    with rf_open(src) as f:
        for line in f:
            yield json.loads(line)


def loading(src: (PATH_TYPE, ...), src_type=None):
    if isinstance(src, PATH_TYPE):
        suffix = pathlib.PurePath(src).suffix[1:]
        if suffix == "csv" or src_type == "csv":
            return load_csv(src)
        elif suffix == "json" or src_type == "json":
            return load_json(src)
        else:
            return load_file(src)
    elif callable(src):
        return src()
    return src



import re

from .process_pattern_base import *

mode_dict = {}

__all__ = ["register", "mode_dict", "init_patterns", "line_init_patterns"]


def register(number):
    def _register(func):
        if number in mode_dict:
            logger.warning(
                "mode-%s %s existed, overriding by %s"
                % (number, mode_dict[number].__name__, func.__name__)
            )
        else:
            logger.info("register mode-%s %s" % (number, func.__name__))
        mode_dict[number] = func
        return func

    return _register


def _init_line(line):
    try:
        line = line.strip()
        line1, line2 = line.split(':')
        return line, (line1, line2)
    except Exception:
        raise ValueError('error in %s' % line)


@register(0)
def _init_patterns0(line, pps=None):
    line, (line1, line2) = _init_line(line)

    try:
        v = line1.strip()
        ps1 = []
        p = re.compile(v)
        ps1.append(p)

        v = line2.strip()
        assert v

        ps2 = []
        p = re.compile(v)
        ps2.append(p)
    except Exception:  # pragma: no cover
        raise ValueError('error in %s' % line)

    if pps is not None and isinstance(pps, dict):
        pps[line] = (ps1, ps2)
    return line, (ps1, ps2)


@register(1)
def _init_patterns1(line, pps):
    line, (line1, line2) = _init_line(line)

    try:
        vs = line1.split()
        ps1 = []
        for v in vs:
            p = re.compile(v)
            ps1.append(p)

        vs = line2.split()
        ps2 = []
        for v in vs:
            p = re.compile(v)
            ps2.append(p)

    except Exception:  # pragma: no cover
        raise ValueError('error in %s' % line)

    if pps is not None and isinstance(pps, dict):
        pps[line] = (ps1, ps2)

    return line, (ps1, ps2)


def _init_patterns(location, patterns):
    """
    创建一个模式字典

    Parameters
    ----------
    location: str
        正则规则文件位置
    patterns:

    Returns
    -------
    dict
        模式字典
    """
    pps = {}
    with open(location) as f:
        for line in f:
            try:
                patterns(line, pps)
            except ValueError as e:  # pragma: no cover
                print(e)
                continue
    return pps


def init_patterns(location, mode):
    return _init_patterns(location, mode_dict[mode])


def line_init_patterns(line, mode, pps=None):
    """

    Parameters
    ----------
    line
    mode: int
    pps

    Returns
    -------

    """
    if mode not in mode_dict:
        raise KeyError(
            "available mode is %s" % list(mode_dict.keys())
        )
    return mode_dict[mode](line, pps)



# coding: utf-8
# 2019/9/20 @ tongshiwei

__all__ = [
    "gitignore",
    "pytest", "coverage",
    "pysetup", "sphinx_conf", "makefile",
    "readthedocs", "travis", "nni",
    "template_copy",
]

import functools
from shutil import copyfile

from longling import wf_open, PATH_TYPE
from longling.lib.path import abs_current_dir, path_append
from longling.lib.process_pattern import default_variable_replace as dvr
from longling.lib.utilog import config_logging, LogLevel

logger = config_logging(logger="arch", console_log_level=LogLevel.INFO)

META = path_append(abs_current_dir(__file__), "meta_docs")
default_variable_replace = functools.partial(dvr, quotation="\'")


def template_copy(src: PATH_TYPE, tar: PATH_TYPE, default_value: (str, dict, None) = "", quotation="\'", **variables):
    """
    Generate the tar file based on the template file where the variables will be replaced.
    Usually, the variable is specified like `$PROJECT` in the template file.


    Parameters
    ----------
    src: template file
    tar: target location
    default_value: the default value
    quotation: the quotation to wrap the variable value
    variables: the real variable values which are used to replace the variable in template file
    """
    with open(src) as f, wf_open(tar) as wf:
        for line in f:
            print(
                default_variable_replace(line, default_value=default_value, quotation=quotation, **variables),
                end='', file=wf
            )


def gitignore(atype: str = "", tar_dir: PATH_TYPE = "./"):
    """

    Parameters
    ----------
    atype: the gitignore type, currently support `docs` and `python`
    tar_dir: target directory

    """
    src = path_append(META, "gitignore", "%s.gitignore" % atype)
    tar = path_append(tar_dir, ".gitignore")
    logger.info("gitignore: copy %s -> %s" % (src, tar))
    copyfile(src, tar)


def pytest(tar_dir: PATH_TYPE = "./"):
    src = path_append(META, "pytest.ini")
    tar = path_append(tar_dir, "pytest.ini")
    logger.info("pytest: copy %s -> %s" % (src, tar))
    copyfile(src, tar)


def coverage(tar_dir: PATH_TYPE = "./", **variables):
    src = path_append(META, "setup.cfg.template")
    tar = path_append(tar_dir, "setup.cfg")
    logger.info("coverage: template %s -> %s" % (src, tar))
    template_copy(src, tar, quotation="", **variables)


def pysetup(tar_dir="./", **variables):
    src = path_append(META, "setup.py.template")
    tar = path_append(tar_dir, "setup.py")
    logger.info("pysetup: template %s -> %s" % (src, tar))
    template_copy(src, tar, **variables)


def sphinx_conf(tar_dir="./", **variables):
    src = path_append(META, "docs/conf.py.template")
    tar = path_append(tar_dir, "conf.py")
    logger.info("sphinx_conf: template %s -> %s" % (src, tar))
    template_copy(src, tar, **variables)

    src = path_append(META, "docs/.math.json")
    tar = path_append(tar_dir, ".math.json")
    logger.info("sphinx_conf: copy %s -> %s" % (src, tar))
    copyfile(src, tar)


def readthedocs(tar_dir="./"):
    src = path_append(META, "docs/.readthedocs.yml")
    tar = path_append(tar_dir, ".readthedocs.yml")
    logger.info("readthedocs: copy %s -> %s" % (src, tar))
    copyfile(src, tar)


def docker():
    pass


def travis(tar_dir: PATH_TYPE = "./"):
    src = path_append(META, ".travis.yml")
    tar = path_append(tar_dir, ".travis.yml")
    logger.info("travis: copy %s -> %s" % (src, tar))
    copyfile(src, tar)


def gitlab_ci():
    pass


def makefile(tar_dir="./", **variables):
    src = path_append(META, "Makefile.template")
    tar = path_append(tar_dir, "Makefile")
    logger.info("makefile: template %s -> %s" % (src, tar))
    template_copy(src, tar, default_value=None, quotation='', **variables)


def nni(tar_dir="./"):
    src_dir = path_append(META, "nni")
    for file in ["config.yml", "search_space.json"]:
        copyfile(path_append(src_dir, file), path_append(tar_dir, file))
